#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "handlers.h"
#include "net.h"


#define CP_LOAD     0
#define CP_FIND     1
#define CP_READ     2
#define CP_WRITE    3
#define CP_SIZE     4
#define CP_OFFSETS  5
#define CP_ALLOC    6
#define CP_FREE     7
#define CP_CALL     8


int loop(struct cpype* pype)
{
    fputs("[+] Entering main loop\n", stderr);
    char cmd;
    while (1)
    {
        if (recv_n(pype->fin, &cmd, sizeof(cmd)))
        {
            break;
        }
        switch (cmd) {
            case CP_LOAD:
                cp_load(pype);
                break;
            case CP_FIND:
                cp_find(pype);
                break;
            case CP_READ:
                cp_read(pype);
                break;
            case CP_WRITE:
                cp_write(pype);
                break;
            case CP_SIZE:
                cp_size(pype);
                break;
            case CP_OFFSETS:
                cp_offsets(pype);
                break;
            case CP_ALLOC:
                cp_alloc(pype);
                break;
            case CP_FREE:
                cp_free(pype);
                break;
            case CP_CALL:
                cp_call(pype);
                break;
            default:
                break;
        }
    }
    return 0;
}

static int init(struct cpype* pype)
{
    char magic[4];
    if (recv_n(pype->fin, &magic, sizeof(magic)))
    {
        fputs("[-] Failed to receive magic\n", stderr);
        return -1;
    }
    if (strncmp(magic, "cpyp", sizeof(magic)))
    {
        fputs("[-] Magic check failed\n", stderr);
        return -1;
    }
    if (send_n(pype->fout, &magic, sizeof(magic)))
    {
        fputs("Failed to send magic\n", stderr);
        return -1;
    }
    uint8_t word_size = sizeof(void*);
    if (send_n(pype->fout, &word_size, sizeof(word_size)))
    {
        fputs("Failed to send word size\n", stderr);
        return -1;
    }
    return 0;
}

int main(int argc, char** argv)
{
    struct cpype pype;
    int rtn = -1;
    if (argc == 2)
    {
        int sock = receive_connection(argv[1]);
        if (sock < 0)
        {
            fputs("[-] Failed to receive connection\n", stderr);
            goto end;
        }
        pype.fin = sock;
        pype.fout = sock;
    }
    else
    {
        pype.fin = 0;
        pype.fout = 1;
    }

    if (init(&pype))
    {
        fputs("[-] Failed to initialize pype\n", stderr);
        goto end;
    }
    rtn = loop(&pype);
end:
    close(pype.fin);
    if (pype.fout != pype.fin)
    {
        close(pype.fout);
    }
    fflush(stderr);
    return rtn;
}
