#include "net.h"

#include <arpa/inet.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <unistd.h>

#include "server.h"


int send_n(int fd, void* data, size_t n)
{
    size_t sent = 0;
    while (sent < n)
    {
        ssize_t r = write(fd, data, n - sent);
        if (r == -1 || r == 0)
        {
            fputs("Failed to write\n", stderr);
            return -1;
        }
        sent += r;
    }
    return 0;
}

int recv_n(int fd, void* data, size_t n)
{
    size_t recvd = 0;
    while (recvd < n)
    {
        ssize_t r = read(fd, data, n - recvd);
        if (r == -1 || r == 0)
        {
            fputs("[-] Failed to read\n", stderr);
            return -1;
        }
        recvd += r;
    }
    return 0;
}

int receive_connection(const char* port_str)
{
    char* endptr;
    long port = strtol(port_str, &endptr, 10);
    int sock_listen = -1;
    int sock_talk = -1;

    if (endptr == port_str || endptr[0])
    {
        fprintf(stderr, "[-] Invalid port string: '%s'\n", port_str);
        goto end;
    }
    sock_listen = socket(AF_INET, SOCK_STREAM, 0);
    if (sock_listen < 0)
    {
        fputs("[-] Failed to make socket\n", stderr);
        goto end;
    }

    // Allow reusing the port immediately after the server exits
    int optval = 1;
    if (setsockopt(sock_listen, SOL_SOCKET, SO_REUSEADDR,
        &optval, sizeof(optval)) < 0)
    {
        fputs("[-] Could not set socket options\n", stderr);
        goto end;
    }

    struct sockaddr_in server_address;
    memset(&server_address, 0, sizeof(server_address));
    server_address.sin_family = AF_INET;
    server_address.sin_port = htons(port);
    if (inet_pton(AF_INET, "0.0.0.0", &(server_address.sin_addr)) <= 0)
    {
        fputs("[-] Invalid address\n", stderr);
        goto end;
    }
    if (bind(sock_listen, (struct sockaddr*)&server_address,
        sizeof(server_address)) < 0)
    {
        fprintf(stderr, "[-] Could not bind to port %ld\n", port);
        goto end;
    }
    if (listen(sock_listen, 1) < 0) {
        fputs("[-] Could not listen on socket\n", stderr);
        goto end;
    }
    fprintf(stderr, "[+] Waiting for connection on port %ld . . .\n",
            port);
    struct sockaddr_in client_address;
    socklen_t client_address_len = sizeof(client_address);
    sock_talk = accept(sock_listen,
                                  (struct sockaddr*)&client_address,
                                  &client_address_len);

    // Close the listening socket after accepting connection
    close(sock_listen);

    if (sock_talk < 0) {
        fputs("[-] Could not accept connection\n", stderr);
        goto end;
    }

    // Print client connection information (optional)
    char client_ip[INET_ADDRSTRLEN];
    inet_ntop(AF_INET, &(client_address.sin_addr), client_ip, INET_ADDRSTRLEN);
    fprintf(stderr, "[+] Connection accepted from %s:%d\n",
            client_ip, ntohs(client_address.sin_port));

end:
    if (sock_listen >= 0)
    {
        close(sock_listen);
    }
    return sock_talk;
}
